/* Predicates used for the computation of both static and dynamic semantics.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `common'
**
**    This module contains predicates that are common to the static
**    and dynamic semantics of SIL and SSFL.
**
**    Predicates defined in this module:
**
**    compose/3, getEnv/3, isEnv/1, isConst/1, isLoc/1, isTerminalAe/2,
**    mask/3, projection/3, tbop/5. 
*/

/*
**    isConst(X)
**
**    True if X is tt, ff or an integer.
*/

isConst(tt).
isConst(ff).
isConst(X) :-
    integer(X).

/*
**    isEnv(X)
**        - succeeds if X is an environment.
**
**    Note: Environments are represented by lists of terms of the form
**          e(X, Y), where X is an identifier and Y is what is associated
**          to S in the environment.
*/

isEnv([_|_]).
isEnv([]).

/*
**    getEnv(E, X, Y)
**        - E represents an environment;
**        - X is an identifier;
**        - Y is any term.
**
**    Succeeds if, in the environment E, identifier X is associated to Y.
*/

getEnv([e(X, Z)|_], X, Y) :- !, Y=Z.
getEnv([_|T], X, Y) :-
    getEnv(T, X, Y).

/*
**    compose(A, B, C)
**        - A, B, C are environments.
**
**    Realizes the composition of environments: compose(A, B, C)
**    succeeds if and only if A[B] = C. Since environments are
**    implemented as ordinary prolog lists, compose/3 is simply the
**    classical append/3 with the first two arguments swapped.
*/

compose(E, [], E).
compose(E1, [B|E2], [B|E3]) :-
    compose(E1, E2, E3).

/*
**    projection(E, V, A)
**       - E is an environment;
**       - V is a set of identifiers (represented by a list);
**       - A is an environment.
**
**    True if A is is obtained from E by discarding all the
**    associations for identifiers that are not present in V.
*/
 
projection([e(X, Y) | E1], V, [e(X, Y) | E2]) :-
    clair_member(X, V),
    !,
    projection(E1, V, E2).
projection([_|T], V, E) :-
    projection(T, V, E).
projection([], _, []).

/*
**    mask(E, V, A)
**       - E is an environment;
**       - V is a set of identifiers (represented by a list);
**       - A is an environment.
**
**    True if A is is obtained from E by keeping only the
**    associations for identifiers that are not present in V.
*/

mask([e(X, _) | E1], V, E2) :-
    clair_member(X, V),
    !,
    mask(E1, V, E2).
mask([e(X, Y) | E1], V, [e(X, Y) | E2]) :-
    mask(E1, V, E2).
mask([], _, []).

/*
**    Definition of the binary operators.
**    -----------------------------------
**
**    tbop(N, T1, T2, T, O)
**       - N  is the abstract syntax name for operator  O;
**       - T1 is the type of the 1st argument;
**       - T2 is the type of the 2nd argument;
**       - T  is the type of the operation's result;
**       - O  is the associated (semantic) operation.
**
**    True if N is a binary operator with domain T1xT2 and codomain
**    T and corresponding to the (semantic) operation O.
*/

/* Arithmetic operators. */

tbop(eq, int, int, bool, =:=).       /* =:= builtin operator */
tbop(ne, int, int, bool, =\=).       /* =\= builtin operator */
tbop(lt, int, int, bool, <).         /* <   builtin operator */
tbop(le, int, int, bool, =<).        /* =<  builtin operator */
tbop(ge, int, int, bool, >=).        /* >=  builtin operator */
tbop(gt, int, int, bool, >).         /* >   builtin operator */

tbop(plus, int, int, int, +).        /* +   builtin operator */
tbop(minus, int, int, int, -).       /* -   builtin operator */
tbop(or, bool, bool, bool, semOr).   /* semOr/3  defined in `runtime' */

tbop(prod, int, int, int, *).        /* *   builtin operator */
tbop(div, int, int, int, //).        /* //  builtin operator */
tbop(mod, int, int, int, mod).       /* mod builtin operator */
tbop(and, bool, bool, bool, semAnd). /* semAnd/3 defined in `runtime' */

/*
**    isTerminalAe(Fl, El)
**        - Fl is a list of formals;
**        - El is a list of actual expressions.
**
**    True if El is a terminal configuration for an actual expression
**    corresponding to the list of formals Fl.  This predicate is used
**    to check the evaluation of actual expressions both for SIL and
**    SFL.
*/

isTerminalAe([ival(_, _)|Form], [C|Ae]) :-
    isConst(C),
    isTerminalAe(Form, Ae).
isTerminalAe([fval(_, _)|Form], [C|Ae]) :-
    isConst(C),
    isTerminalAe(Form, Ae).
isTerminalAe([name(_, _)|Form], [E|Ae]) :-
    eFV(E, []),
    isTerminalAe(Form, Ae).
isTerminalAe([ref(_, _)|Form], [C|Ae]) :- 
    (isLoc(C) ; C=a(_, _)),
    isTerminalAe(Form, Ae).
isTerminalAe([const(_, _)|Form], [C|Ae]) :-
    isConst(C),
    isTerminalAe(Form, Ae).
isTerminalAe([copy(_, _)|Form], [lr(_, _)|Ae]) :-
    isTerminalAe(Form, Ae).
isTerminalAe([funcpar(_, _, _)|Form], [func(_, _)|Ae]) :-
    isTerminalAe(Form, Ae).
isTerminalAe([procpar(_, _)|Form], [proc(_, _)|Ae]) :-
    isTerminalAe(Form, Ae).
isTerminalAe([], []).

isLoc(L) :-
    integer(L).
