/* Predicates for error-handling.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `errors'
**
**    This module contains predicates to handle the erros
**    found by the checks derived from the static semantics.
**    The main predicate is error/1;  this asserts the clause
**    `scheck(failed)' and notifies the error to the user.
**
**    Predicates defined in this module:
**
**    error/1, n/0, notify/1, w/1, wform/1, wfrm/1, wid/1, wlimits/1,
**    wop/1, wstype/1, wtype/1. 
*/


/*
**    error(X)
**        - X is a list of atom identifying the error.
**
**    Asserts the clause `scheck(failed)' and notifies the error
**    to the user.
*/

error(X) :-
  setGlobal(scheck(failed)),
  notify(X).

/*
**    Clauses for the generation of error messages.
**    ---------------------------------------------
*/

/*
**    Abbreviations for write/1 and nl/0.
*/

w(X) :- write(X).
n :- nl.

/*
**    Visualization of operators.
*/

wop(not) :- w('`not''').
wop(uminus) :- w('unary `-''').

wop(eq) :- w('`==''').
wop(ne) :- w('`<>''').
wop(lt) :- w('`<''').
wop(le) :- w('`<=''').
wop(ge) :- w('`>=''').
wop(gt) :- w('`>''').

wop(plus) :- w('`+''').
wop(minus) :- w('`-''').
wop(or) :- w('`or''').

wop(prod) :- w('`*''').
wop(div) :- w('`div''').
wop(mod) :- w('`mod''').
wop(and) :- w('`and''').

/*
**    Visualization of identifiers.
*/

wid(X) :- w('`'), w(X), w('''').

/*
**    Visualization of types.
*/

wstype(int) :- w('integer').
wstype(bool) :- w('boolean').

wtype(X) :- var(X), w('`?''').
wtype(l(T)) :- w('`(loc) '), wstype(T), w(''''). 
wtype(int) :- w('`integer''').
wtype(bool) :- w('`boolean''').
wtype(a(L, T)) :- w('`array ['), wlimits(L), w('] of '),
    wstype(T), w('''').
wtype(ftype(Ft, T)) :- w('`function('), wform(Ft), w(') -> '),
    wstype(T), w('''').
wtype(ptype(Ft)) :- w('`procedure('), wform(Ft), w(')''').

wlimits([b(L, H)]) :- w(L), w('..'), w(H).
wlimits([First|Rest]) :- wlimits([First]), w(','), wlimits(Rest).
wlimits([]).

/*
**    Visualization of formal parameters.
*/

wform([F]) :- wfrm(F).
wform([First|Rest]) :- wform([First]), w(','), wform(Rest).
wform([]).

wfrm(ival(_, T)) :- wstype(T).
wfrm(fval(_, T)) :- wstype(T).
wfrm(name(_, T)) :- w('(name) '), wstype(T).
wfrm(ref(_, T)) :- w('(ref) '), wstype(T).
wfrm(const(_, T)) :- w('(const) '), wstype(T).
wfrm(copy(_, T)) :- w('(copy) '), wstype(T).
wfrm(funcpar(_, F, T)) :-
    w('function('), wform(F), w('):'), wstype(T).
wfrm(procpar(_, F)) :-
    w('procedure('), wform(F), w(')').

/*
**    Notification of error messages.
*/

notify([nolvalue, ass, Identifier]) :-
    w('No L-value: cannot assign to '), wid(Identifier), n.

notify([nolvalue, callByReference, X]) :-
    w('No L-value: formal '), wid(X),
    w(' must be passed by reference'), n.

notify([nolvalue, callByValueResult, X]) :-
    w('No L-value: formal '), wid(X),
    w(' must be passed by value-result'), n.

notify([undeclared, id, Identifier]) :-
    w('Undeclared identifier: '), wid(Identifier), n.

notify([undeclared, var, Identifier]) :-
    w('Undeclared variable: '), wid(Identifier), n.

notify([undeclared, array, Array]) :-
    w('Undeclared array: '), wid(Array), n.

notify([undeclared, function, Func]) :-
    w('Undeclared function: '), wid(Func), n.

notify([undeclared, procedure, Proc]) :-
    w('Undeclared procedure: '), wid(Proc), n.

notify([typeMismatch, arrayIndexType, Array]) :-
    w('Type mismatch: bad indexes for array '), wid(Array),
    w(', must all be `integer'''), n.

notify([typeMismatch, arrayIndexNumber, Array]) :-
    w('Type mismatch: wrong number of indexes for array '),
    wid(Array), n.

notify([typeMismatch, while, Type]) :-
    w('Type mismatch: `while E do C'', E must be `boolean'''), n,
    w('               '), wtype(Type), w(' is incorrect'), n.

notify([typeMismatch, cif, Type]) :-
    w('Type mismatch: `if E then C1 else C2'', E must be `boolean'''), n,
    w('               '), wtype(Type), w(' is incorrect'), n.

notify([typeMismatch, ass, Identifier, EType, RType]) :-
    w('Type mismatch: a value of type '), wtype(RType),
    w(' cannot be assigned to'), n,
    w('               variable '), wid(Identifier),
    w(' ('), wtype(EType), w(')'), n.

notify([typeMismatch, formal, Identifier, EType, RType]) :-
    w('Type mismatch: a value of type '), wtype(RType),
    w(' cannot be passed,'), n,
    w('               being formal '), wid(Identifier),
    w(' of type '), wtype(EType), n.

notify([typeMismatch, function, wrongNumberOfActuals, F]) :-
    w('Type mismatch: wrong number of actual parameters'), n,
    w('               for function '), wid(F), n.

notify([typeMismatch, procedure, wrongNumberOfActuals, P]) :-
    w('Type mismatch: wrong number of actual parameters'), n,
    w('               for procedure '), wid(P), n.

notify([typeMismatch, funDecl, F, FType, RType]) :-
    w('Type mismatch: function '), wid(F),
    w(' must return '), wtype(FType), n,
    w('               body expression type cannot be '),
    wtype(RType), n.

notify([typeMismatch, if, Type]) :-
    w('Type mismatch: `if E then E1 else E2'', E must be `boolean'''), n,
    w('               '), wtype(Type), w(' is incorrect'), n.

notify([typeMismatch, if, T1, T2]) :-
    w('Type mismatch: `if E then E1 else E2'', E1 and E2 not of the'),
    w(' same type,'), n,
    w('               cannot mix '), wtype(T1), w(' and '),
    wtype(T2), n.

notify([typeMismatch, id, Identifier, EType, RType]) :-
    w('Type mismatch: '), wid(Identifier),
    w(' is '), wtype(RType), n,
    w('               where needed to be '),
    wtype(EType), n.

notify([formalMismatch, function, FormN, FormF, ActN, ActF]) :-
    w('Functional parameter mismatch: formal parameters must be'), n,
    w('  equal in number, type and mechanism. Formal and actual are:'), n,
    w('    function '), wid(FormN), w(' takes '), wform(FormF), n,
    w('    function '), wid(ActN),  w(' takes '), wform(ActF), n.

notify([formalMismatch, procedure, FormN, FormF, ActN, ActF]) :-
    w('Procedural parameter mismatch: formal parameters must be'), n,
    w('  equal in number, type and mechanism. Formal and actual are:'), n,
    w('    procedure '), wid(FormN), w(' takes '), wform(FormF), n,
    w('    procedure '), wid(ActN),  w(' takes '), wform(ActF), n.

notify([typeMismatch, notAFunction, expr, N]) :-
    w('Type mismatch: formal '), wid(N),
    w(' is a function, cannot pass anything else'), n.

notify([typeMismatch, notAProcedure, expr, N]) :-
    w('Type mismatch: formal '), wid(N),
    w(' is a procedure, cannot pass anything else'), n.

notify([typeMismatch, notAFunction, id, N, X, Dx]) :-
    w('Type mismatch: formal '), wid(N),
    w(' is a function, cannot pass'), n,
    w('               '), wid(X), w(' ('), wtype(Dx), w(')'), n.

notify([typeMismatch, notAProcedure, id, N, X, Dx]) :-
    w('Type mismatch: formal '), wid(N),
    w(' is a procedure, cannot pass'), n,
    w('               '), wid(X), w(' ('), wtype(Dx), w(')'), n.

notify([typeMismatch, functionReturn, N, EType, R, RType]) :-
    w('Type mismatch: formal '), wid(N),
    w(' is function returning '), wtype(EType), n,
    w('               cannot pass '), wid(R), w(' (returns '),
    wstype(RType), w(')'), n.
 
notify([typeMismatch, array, Array, BType, RType]) :-
    w('Type mismatch: '), wid(Array),
    w(' is '), wtype(RType), w(' where needed to be'), n,
    w('               array [...] of '), wtype(BType), n.

notify([typeMismatch, uop, Uop, EType, RType]) :-
    w('Type mismatch: '), wop(Uop), w(' applied to '),
    wtype(RType), w(' instead of '), wtype(EType), n.

notify([typeMismatch, bop, Bop, TArg1, TArg2]) :-
    w('Type mismatch: '), wop(Bop), w(' applied to '),
    wtype(TArg1), w(' and '), wtype(TArg2), n.

notify([typeMismatch, varDecl, Identifier, EType, RType]) :-
    w('Type mismatch: variable '), wid(Identifier),
    w(' defined '), wtype(EType), w(' and initialized with '),
    wtype(RType), n.
 
notify([typeMismatch, constDecl, Identifier, EType, RType]) :-
    w('Type mismatch: constant '), wid(Identifier),
    w(' defined '), wtype(EType), w(' and initialized with '),
    wtype(RType), n.

notify([typeMismatch, arrayDecl, Identifier, EType, RType]) :-
    w('Type mismatch: array '), wid(Identifier),
    w(' defined [...] of '), wtype(EType), n,
    w('               and initialized with '),
    wtype(RType), n.

notify([malformed, arrayDecl, Identifier]) :-
    w('Malformed declaration: array '), wid(Identifier),
    w(', `L..H'' requires L <= H'), n.

notify([malformed, procDecl, Proc, Duplicates]) :-
    w('Malformed declaration: procedure '), wid(Proc),
    w(', duplicated formal'), n,
    w('                       parameters '), w(Duplicates), n.

notify([malformed, funDecl, Fun, Duplicates]) :-
    w('Malformed declaration: function '), wid(Fun),
    w(', duplicated formal'), n,
    w('                       parameters '), w(Duplicates), n.

notify([malformed, parallel, Duplicates]) :-
    w('Malformed declaration: composite declaration `D1 | D2'' '),
    w('requires D1 and D2'), n,
    w('                       to be disjoint, duplicate variables are'),
    n,
    w('                       '), w(Duplicates), n.
