/* Implementation of the dynamic semantics for the functional language.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `fdinsem'
**
**    This module contains the predicates implementing the dynamic
**    semantics of SFL.
**
**    Predicates defined in this module:
**
**    ae1t/4, d1t/3, e1t/3, expTrace/3, traceDependentClause/2. 
*/

:- multifile(traceDependentClause/2).

/*
**    Clauses for the dynamic semantics of declarations.
**    ==================================================
**
**    d1t(E, D, D1)
**        - E     is a value environment;
**        - D, D1 are declarations.
**
**    True if, in the environment E, the declaration configuration
**    D transitions into D1 in a number of steps.
**
**    N.B. When the tracer is active, the transition D -> D1
**         occurs in one step; otherwise, it occurs in as many steps
**         as necessary to ensure that D1 is an environment (terminal
**         configuration for declarations).
*/

/*
**    Nil.
**    ----
*/

d1t(_, nil, []).

/*
**    Simple definition.
**    ------------------
*/

d1t(_, def(X, _, Y), [e(X, Y)]) :-
    isConst(Y).
d1t(E, def(X, T, Y), def(X, T, Y1)) :-
    expStar(E, Y, Y1).

/*
**    Sequential composition.
**    -----------------------
*/

d1t(_, seq(X, Y), Z) :-
    isEnv(X), isEnv(Y),
    compose(X, Y, Z).
d1t(E, seq(X, Y), seq(X, Y1)) :-
    isEnv(X),
    compose(E, X, E1),
    declStar(E1, Y, Y1).
d1t(E, seq(X, Y), seq(X1, Y)) :-
    declStar(E, X, X1).

/* 
**    Private composition.
**    --------------------
*/

d1t(_, private(X, Y), Y) :-
    isEnv(X), isEnv(Y).
d1t(E, private(X, Y), private(X, Y1)) :-
    isEnv(X),
    compose(E, X, E1),
    declStar(E1, Y, Y1).
d1t(E, private(X, Y), private(X1, Y)) :-
    declStar(E, X, X1).

/*
**    Parallel composition.
**    ---------------------
*/

d1t(_, parallel(X, Y), Z) :-
    isEnv(X), isEnv(Y),
    compose(X, Y, Z).
d1t(E, parallel(X, Y), parallel(X, Y1)) :-
    isEnv(X),
    declStar(E, Y, Y1).
d1t(E, parallel(X, Y), parallel(X1, Y)) :-
    declStar(E, X, X1).

/*
**    Function definition.
**    --------------------
*/

d1t(E, fundef(F, Formals, _, Body), [e(F, func(Formals, NewBody))]) :-
    eFV(Body, FVbody), formDV(Formals, DVformals),
    subtraction(FVbody, DVformals, V),
    projection(E, V, E1),
    (E1 = [] -> NewBody = Body ; NewBody = let(E1, Body)).

/*
**    Recursive definition.
**    ---------------------
*/

d1t(_, rec(E0), E1) :-
    isEnv(E0),
    makeRecEnv(E0, E0, E1).
d1t(E, rec(D), rec(D1)) :-
    dDV(D, DVd), dFV(D, FVd),
    clair_intersection(DVd, FVd, V0),
    mask(E, V0, E0),
    declStar(E0, D, D1).

/*
**    Parameter passing.
**    ------------------
*/

d1t(_, parPass(Formals, Ae), R) :-
    isTerminalAe(Formals, Ae), 
    makeEnv(Formals, Ae, R).
d1t(E, parPass(Formals, Ae), parPass(Formals, Ae1)) :- 
    ae1t(E, Formals, Ae, Ae1).

/*
**    Clauses for the dynamic semantics of expressions.
**    =================================================
**
**    e1t(A, E, E1)
**        - A     is a value environment;
**        - E, E1 are expressions.
**
**    True if, in the environment A, the expression configuration E
**    transitions into E1 in a number of steps.
**
**    See above for details about the actual number of steps
**    performed.
*/

/*
**    Boolean negation.
**    -----------------
*/

e1t(_, not(tt), ff).
e1t(_, not(ff), tt).
e1t(E, not(X), not(X1)) :-
    expStar(E, X, X1).

/*
**    Unary minus.
**    ------------
*/

e1t(_, uminus(X), Y) :-
    integer(X),
    Y is -X.
e1t(E, uminus(X), uminus(X1)) :-
    expStar(E, X, X1).

/*
**    Identifier.
**    -----------
*/

e1t(E, i(X), Y) :-
    getEnv(E, X, Y).

/*
**    Binary operators.
**    -----------------
**
**    The following clauses are indeed a bit complicated.  However,
**    together with the ones in `statsem' and `common', they serve the
**    purpose to avoid a proliferation of clauses for the treatment of
**    binary operators.  An advantage is that adding a new operator is
**    very easy and requires only few changes: it suffices to add a
**    new clause for tbop/5 (in `common') and, if necessary, the
**    definition of the associated semantic operator in `runtime'.
*/

e1t(_, X, Y) :-
    X =.. [O, Arg1, Arg2],
    tbop(O, D, D, R, F),
    isConst(Arg1), isConst(Arg2),
    (D = int ->                   /* ==> arithmetic or relational op */
        (Exp =.. [F, Arg1, Arg2],
        (R = int ->               /* ==> arithmetic op */
            (Eval =.. [is, Y, Exp], call(Eval))
        ;                         /* ==> relational op */
            (call(Exp) -> Y = tt ; Y = ff)))
    ;                             /* ==> boolean op */
        (Eval =.. [F, Arg1, Arg2, Y], call(Eval))).

e1t(E, X, Y) :-
    X =.. [O, A, B],
    tbop(O, _, _, _, _),
    isConst(A),
    expStar(E, B, B1),
    Y =.. [O, A, B1].

e1t(E, X, Y) :-
    X =.. [O, A, B],
    tbop(O, _, _, _, _),
    expStar(E, A, A1),
    Y =.. [O, A1, B].

/*
**    Conditional expression.
**    -----------------------
*/

e1t(_, if(tt, Y, _), Y).
e1t(_, if(ff, _, Z), Z).
e1t(E, if(X, Y, Z), if(X1, Y, Z)) :-
    expStar(E, X, X1).

/*
**    Expression block.
**    -----------------
*/

e1t(_, let(X, Y), Y) :-
    isEnv(X),
    isConst(Y).
e1t(E, let(X, Y), let(X, Y1)) :-
    isEnv(X),
    compose(E, X, E1),
    expStar(E1, Y, Y1).
e1t(E, let(X, Y), let(X1, Y)) :-
    declStar(E, X, X1).

/*
**    Function call.
**    --------------
*/

e1t(E, funcall(F, Ae), let(parPass(Formals, Ae), Body)) :-
    getEnv(E, F, func(Formals, Body)).

/*
**    Input.
**    ------
*/

e1t(_, input(X), C) :-
    doInput(X, C).

/*
**    Clauses for the dynamic semantics of actual expressions.
**    ========================================================
**
**    ae1t(E, Fl, El, El1)
**        - E       is a value environment;
**        - Fl      is a list of formal parameters;
**        - El, El1 are expressions attuali.
**
**    True if, in the environment E and with the list of formals Fl,
**    the actual expression configuration El transitions into El1 in a
**    number of steps.
**
**    See above for details about the actual number of steps
**    performed.
*/

/*
**    Call by value.
**    --------------
*/

ae1t(E, [fval(_, _) | Form], [C | Ae], [C | Ae1]) :-
    isConst(C), 
    ae1t(E, Form, Ae, Ae1).
ae1t(E, [fval(_, _) | _], [C | Ae], [C1 | Ae]) :-
    expStar(E, C, C1).

/*
**    Call by name.
**    -------------
*/

ae1t(E, [name(_, _) | Form], [Exp | Ae], [NewExp | Ae1]) :-
    eFV(Exp, V),
    (V=[] ->
        NewExp=Exp
    ;
        (projection(E, V, E0), NewExp=let(E0, Exp))),
    ae1t(E, Form, Ae, Ae1).

/*
**    Functional parameters.
**    ----------------------
*/

ae1t(E, [funcpar(_, _, _)|Form], [func(F, B)|Ae], [func(F, B)|Ae1]) :-
    ae1t(E, Form, Ae, Ae1).
ae1t(E, [funcpar(_, _, _)|Form], [i(FuncId)|Ae], [func(F, B)|Ae1]) :-
    getEnv(E, FuncId, func(F, B)),
    ae1t(E, Form, Ae, Ae1).

ae1t(_, [], [], []).

/*
**    expTrace(A, E, Ef)
**        - A     is a value environment;
**        - E, Ef are expressions.
**
**    True if, in the environment A, the expression configuration
**    E transitions into the (terminal) configuration Ef (a constant).
**    As a side-effect, the evaluation of the expression is traced.
*/

expTrace(_, X, X) :-
    isConst(X), !,
    outputConfig(X),
    (\+ traceFile(user) ->
       (write(X), nl)     /* Write the result. */
    ;
       true).

expTrace(E, X, Y) :-
    outputConfig(X),
    e1t(E, X, X1), !, expTrace(E, X1, Y).

/*
**    Clauses of dynamic semantics that depend on the tracer status.
**    ==============================================================
*/

/*
**    expStar(A, E, E1)
**        - A     is a value environment;
**        - E, E1 are expressions.
**
**    True if, in the environment A, the expression configuration E
**    transitions into the configuration E1.
**    If the tracer is active, expStar/3 is equivalent to e1t/3;
**    otherwise expStar/3 invokes itself recursively until E1 is a
**    constant.
*/

traceDependentClause(on, (
    expStar(E, X, Y) :- e1t(E, X, Y), !
)).

traceDependentClause(off, (
    expStar(_, X, X) :- isConst(X), !
)).

traceDependentClause(off, (
    expStar(E, X, Y) :-
        e1t(E, X, X1), !,
        expStar(E, X1, Y)
)).

/*
**    declStar(E, D, D1)
**        - E     is a value environment;
**        - D, D1 are declarations.
**
**    True if, in the environment E, D transitions into D1.
**    If the tracer is active, declStar/3 is equivalent to d1t/3;
**    otherwise declStar/3 invokes itself recursively until D1 is an
**    environment.
*/

traceDependentClause(on, (
    declStar(E, X, Y) :- d1t(E, X, Y), !
)).

traceDependentClause(off, (
    declStar(_, X, X) :- isEnv(X), !
)).

traceDependentClause(off, (
    declStar(E, X, Y) :-
        d1t(E, X, X1), !,
        declStar(E, X1, Y)
)).

/*
**    topDynamicFunctional(E)
**        - E is an expression.
**
**    True if there exists a (finite) sequence of transitions that, in
**    the empty environment, brings E to a terminal configuration
**    consisting of a constant.
**
*/

traceDependentClause(off, (
    topDynamicFunctional(X) :-
        expStar([], X, Y), write(Y), nl
)).

traceDependentClause(on, (
    topDynamicFunctional(X) :-
        expTrace([], X, _)
)).
