/* Implementation of the dynamic semantics for the imperative language.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `idinsem'
**
**    This module contains the predicates implementing the dynamic
**    semantics of SIL.
**
**    Predicates defined in this module:
**
**    ae1t/6, c1t/5, comTrace/5, d1t/5, e1t/5, el1t/5,
**    isConstList/1, traceDependentClause/2.
*/

:- multifile(traceDependentClause/2).

/*
**    Clauses for the dynamic semantics of declarations.
**    ==================================================
**
**    d1t(E, D, S, D1, S1)
**        - E     is a value environment;
**        - D, D1 are declarations;
**        - S, S1 are stores.
**
**    True if, in the environment E, the declaration configuration
**    <D, S> transitions into <D1, S1> in a number of steps.
**
**    N.B. When the tracer is active, the transition <D, S> -> <D1, S1>
**         occurs in one step; otherwise, it occurs in as many steps
**         as necessary to ensure that D1 is an environment (so that
**         <D1, S1> is a terminal configuration for declarations).
*/

/*
**    Nil.
**    ----
*/

d1t(_, nil, S, [], S).

/*    
**    Constant declaration.
**    ---------------------
*/

d1t(_, con(X, T, Y), S, [e(X, d(T, Y))], S) :-
    isConst(Y).
d1t(E, con(X, T, Y), S, con(X, T, Y1), S1) :-
    expStar(E, Y, S, Y1, S1).

/*    
**    Variable declaration (including arrays).
**    ----------------------------------------
*/

d1t(_, var(X, array(Limits, T), Y), S,
    [e(X, d(array(Limits, T), a(Limits, Base)))], S1) :-
    isConst(Y),
    makeArray(Limits, Base, Y, S2),
    append(S, S2, S1).
d1t(E, var(X, array(Limits, T), Y), S, var(X, array(Limits, T), Y1), S1) :-
    expStar(E, Y, S, Y1, S1).

d1t(_, var(X, T, Y), S, [e(X, d(T, l(L)))], [s(L, Y)|S]) :-
    isConst(Y),
    getNewLoc(L).
d1t(E, var(X, T, Y), S, var(X, T, Y1), S1) :-
    expStar(E, Y, S, Y1, S1).

/*
**    Sequential composition.
**    -----------------------
*/

d1t(_, seq(X, Y), S, Z, S) :-
    isEnv(X), isEnv(Y),
    compose(X, Y, Z).
d1t(E, seq(X, Y), S, seq(X, Y1), S1) :-
    isEnv(X),
    compose(E, X, E1),
    declStar(E1, Y, S, Y1, S1).
d1t(E, seq(X, Y), S, seq(X1, Y), S1) :-
    declStar(E, X, S, X1, S1).

/*    
**    Private composition.
**    --------------------
*/

d1t(_, private(X, Y), S, Y, S) :-
    isEnv(X),
    isEnv(Y).
d1t(E, private(X, Y), S, private(X, Y1), S1) :-
    isEnv(X),
    compose(E, X, E1),
    declStar(E1, Y, S, Y1, S1).
d1t(E, private(X, Y), S, private(X1, Y), S1) :-
    declStar(E, X, S, X1, S1).

/*    
**    Parallel composition.
**    ---------------------
*/

d1t(_, parallel(X, Y), S, Z, S) :-
    isEnv(X), isEnv(Y),
    compose(X, Y, Z).
d1t(E, parallel(X, Y), S, parallel(X, Y1), S1) :-
    isEnv(X),
    declStar(E, Y, S, Y1, S1).
d1t(E, parallel(X, Y), S, parallel(X1, Y), S1) :-
    declStar(E, X, S, X1, S1).

/*
**    Procedure declaration.
**    ----------------------
*/

d1t(E, procdef(P, Form, Body), S,
    [e(P, d(proc(Form), proc(Form, NewBody)))], S) :-
    cFV(Body, FVbody), formDV(Form, DVform),
    subtraction(FVbody, DVform, V),
    projection(E, V, E1),
    (E1 = [] ->
        NewBody = Body
    ;
        NewBody = closureC(E1, Body)).

/*
**    Function declaration.
**    ---------------------
*/

d1t(E, fundef(F, Form, T, Body), S,
    [e(F, d(func(Form, T), func(Form, NewBody)))], S) :-
    eFV(Body, FVbody), formDV(Form, DVform),
    subtraction(FVbody, DVform, V),
    projection(E, V, E1),
    (E1 = [] ->
        NewBody = Body
    ;
        NewBody = closureE(E1, Body)).

/*
**    Parameter passing.
**    ------------------
*/

d1t(_, parPass(Formals, Ae), S, Env, S1) :-
    isTerminalAe(Formals, Ae),
    makeEnvStore(Formals, Ae, Env, ParamStore),
    append(ParamStore, S, S1).

d1t(E, parPass(Formals, Ae), S, parPass(Formals, Ae1), S1) :-
    ae1t(E, Formals, Ae, S, Ae1, S1).

/*
**    Recursive declaration.
**    ----------------------
*/

d1t(_, rec(E0), S, E1, S) :-
    isEnv(E0),
    makeRecEnv(E0, E0, E1).
d1t(E, rec(D), S, rec(D1), S1) :-
    dDV(D, DVd), dFV(D, FVd),
    clair_intersection(DVd, FVd, V0),
    mask(E, V0, E0),
    declStar(E0, D, S, D1, S1).

/*
**    Clauses for the dynamic semantics of commands.
**    ==============================================
**
**    c1t(E, C, S, C1, S1)
**        - E     is a value environment;
**        - C, C1 are commands;
**        - S, S1 are stores.
**
**    True if, in the environment E, the command configuration <C, S>
**    transitions into <C1, S1> in a number of steps.
**
**    See above for details about the actual number of steps
**    performed.
*/

/*
**    Nop.
**    ----
*/

c1t(_, nop, S, halt, S).

/*
**    Assignment.
**    -----------
*/

c1t(E, ass(aref(X, El), Y), S, halt, S1) :-
    isConstList(El), isConst(Y),
    getEnvV(E, X, a(Limits, Base)),
    arrayIndex(Limits, El, Offset),
    Loc is Base+Offset,
    changeStore(S, Loc, Y, S1).

c1t(E, ass(aref(X, El), Y), S, ass(aref(X, El), C), S1) :-
    isConstList(El),
    expStar(E, Y, S, C, S1).

c1t(E, ass(aref(X, El), Y), S, ass(aref(X, El1), Y), S1) :-
    el1t(E, El, S, El1, S1).

c1t(E, ass(i(X), Y), S, halt, S1) :-
    isConst(Y),
    getEnvV(E, X, H),
    (H=l(L) ; H=r(L) ; H=c(_, L)),
    changeStore(S, L, Y, S1).

c1t(E, ass(i(X), Y), S, ass(i(X), C), S1) :-
    expStar(E, Y, S, C, S1).

/*
**    Command sequence.
**    -----------------
*/

c1t(_, seq(halt, C2), S, C2, S).
c1t(E, seq(C1, C2), S, seq(C11, C2), S1) :-
    comStar(E, C1, S, C11, S1).

/*
**    Conditional command.
**    --------------------
*/

c1t(_, if(tt, C1, _), S, C1, S).
c1t(_, if(ff, _, C2), S, C2, S).
c1t(E, if(Exp, C1, C2), S, if(B, C1, C2), S1) :-
    expStar(E, Exp, S, B, S1).

/*
**    While command.
**    --------------
*/

c1t(_, while(Exp, C), S, if(Exp, seq(C, while(Exp, C)), halt), S).

/*
**    Command block.
**    --------------
*/

c1t(_, block(D, halt), S, halt, S1) :-
    dispose(D, S, S1).
c1t(E, block(D, C), S, block(D, C1), S1) :-
    isEnv(D),
    compose(E, D, G),
    comStar(G, C, S, C1, S1).
c1t(E, block(D, C), S, block(D1, C), S1) :-
    declStar(E, D, S, D1, S1).

/*
**    Closure block for procedural abstractions.
**    ------------------------------------------
**
**    It is similar to the command block, with the following
**    differences:
**
**        1) it always receives an environment;
**        2) it does not dispose variables.
*/

c1t(_, closureC(_, halt), S, halt, S).
c1t(E, closureC(D, C), S, closureC(D, C1), S1) :-
    compose(E, D, G),
    comStar(G, C, S, C1, S1).

/*
**    Print command.
**    --------------
*/

c1t(_, print(Format, El), S, halt, S) :-
    isConstList(El),
    doPrint(Format, El).

c1t(E, print(Format, El), S, print(Format, El1), S1) :-
    el1t(E, El, S, El1, S1).

/*
**    Procedure call.
**    ---------------
*/

c1t(E, proccall(P, Ae), S, block(parPass(Formals, Ae), Body), S) :-
    getEnvV(E, P, proc(Formals, Body)).

/*
**    Clauses for the dynamic semantics of expressions.
**    =================================================
**
**    e1t(A, E, S, E1, S1)
**        - A     is a value environment;
**        - E, E1 are expressions;
**        - S, S1 are stores.
**
**    True if in the environment A the expression configuration <E, S>
**    transitions into <E1, S1> in a number of steps.
**
**    See above for details about the actual number of steps
**    performed.
*/

/*
**    Boolean negation.
**    -----------------
*/

e1t(_, not(tt), S, ff, S).
e1t(_, not(ff), S, tt, S).
e1t(E, not(X), S, not(X1), S1) :-
    expStar(E, X, S, X1, S1).

/*
**    Unary minus.
**    ------------
*/

e1t(_, uminus(X), S, Y, S) :-
    integer(X),
    Y is -X.
e1t(E, uminus(X), S, uminus(X1), S1) :-
    expStar(E, X, S, X1, S1).

/*
**    Array reference.
**    ----------------
*/

e1t(E, aref(X, El), S, C, S) :-
    isConstList(El),
    getEnvV(E, X, a(Limits, Base)), arrayIndex(Limits, El, Offset),
    Loc is Base+Offset,
    getStore(S, Loc, C).
e1t(E, aref(X, El), S, aref(X, El1), S1) :-
    el1t(E, El, S, El1, S1).

/*
**    Identifier.
**    -----------
*/

e1t(E, i(X), S, C, S) :-
    getEnvV(E, X, D),
    (((D=l(L) ; D=r(L) ; D=c(_, L)), getStore(S, L, C)) ; C=D).

/*
**    Binary operators.
**    -----------------
**
**    The following clauses are indeed a bit complicated.  However,
**    together with the ones in `statsem' and `common', they serve the
**    purpose to avoid a proliferation of clauses for the treatment of
**    binary operators.  An advantage is that adding a new operator is
**    very easy and requires only few changes: it suffices to add a
**    new clause for tbop/5 (in `common') and, if necessary, the
**    definition of the associated semantic operator in `runtime'.
*/

e1t(_, X, S, Y, S) :-
    X =.. [O, Arg1, Arg2],
    tbop(O, D, D, R, F),
    isConst(Arg1), isConst(Arg2),
    (D = int ->                   /* ==> arithmetic or relational op */
        (Exp =.. [F, Arg1, Arg2],
        (R = int ->               /* ==> arithmetic op */
            (Eval =.. [is, Y, Exp], call(Eval))
        ;                         /* ==> relational op */
            (call(Exp) -> Y = tt ; Y = ff)))
    ;                             /* ==> boolean op */
        (Eval =.. [F, Arg1, Arg2, Y], call(Eval))).

e1t(E, X, S, Y, S1) :-
    X =.. [O, A, B],
    tbop(O, _, _, _, _),
    isConst(A),
    expStar(E, B, S, B1, S1),
    Y =.. [O, A, B1].

e1t(E, X, S, Y, S1) :-
    X =.. [O, A, B],
    tbop(O, _, _, _, _),
    expStar(E, A, S, A1, S1),
    Y =.. [O, A1, B].

/*
**    Conditional expression.
**    -----------------------
*/

e1t(_, if(tt, Y, _), S, Y, S).
e1t(_, if(ff, _, Z), S, Z, S).
e1t(E, if(X, Y, Z), S, if(C, Y, Z), S1) :-
    expStar(E, X, S, C, S1).

/*
**    Input.
**    ------
*/

e1t(_, input(X), S, C, S) :-
    doInput(X, C).

/*
**    Expression with side-effects.
**    -----------------------------
*/

e1t(_, side(D, halt, K), S, K, S1) :-
    isConst(K),
    dispose(D, S, S1).
e1t(E, side(D, halt, Exp), S, side(D, halt, Exp1), S1) :-
    compose(E, D, G),
    expStar(G, Exp, S, Exp1, S1).
e1t(E, side(D, C, Exp), S, side(D, C1, Exp), S1) :-
    isEnv(D),
    compose(E, D, G),
    comStar(G, C, S, C1, S1).
e1t(E, side(D, C, Exp), S, side(D1, C, Exp), S1) :-
    declStar(E, D, S, D1, S1).

/*
**    Function call.
**    --------------
*/

e1t(E, funcall(F, Ae), S, let(parPass(Formals, Ae), Body), S) :-
    getEnvV(E, F, func(Formals, Body)).

/*
**    Expression block.
**    -----------------
*/

e1t(_, let(D, K), S, K, S1) :-
    isEnv(D),
    isConst(K),
    dispose(D, S, S1).
e1t(E, let(D, Exp), S, let(D, Exp1), S1) :-
    isEnv(D),
    compose(E, D, G),
    expStar(G, Exp, S, Exp1, S1).
e1t(E, let(D, Exp), S, let(D1, Exp), S1) :-
    declStar(E, D, S, D1, S1).

/*
**    Closure block for expression abstractions.
**    ------------------------------------------
**
**    It is similar to the expression block, with the following
**    differences:
**
**        1) it always receives an environment;
**        2) it does not dispose variables.
*/

e1t(_, closureE(_, K), S, K, S) :-
    isConst(K).
e1t(E, closureE(D, Exp), S, closureE(D, Exp1), S1) :-
    compose(E, D, G),
    expStar(G, Exp, S, Exp1, S1).

/*
**    Clauses for the dynamic semantics of lists of expressions.
**    ==========================================================
**
**    el1t(E, El, S, El1, S1)
**        - E       is a value environment;
**        - El, El1 are lists of expressions;
**        - S, S1   are stores.
**
**    True if, in the environment E, the expression list configuration
**    <El, S> transitions into <El1, S1> in a number of steps.
**
**    See above for details about the actual number of steps
**    performed.
*/

el1t(E, [C | Ae], S, [C | Ae1], S1) :-
    isConst(C),
    el1t(E, Ae, S, Ae1, S1).
el1t(E, [H | Ae], S, [H1 | Ae], S1) :-
    expStar(E, H, S, H1, S1).
el1t(_, [], S, [], S).

/*
**    Clauses for the dynamic semantics of actual expressions.
**    =========================================================
**
**    ae1t(E, Fl, El, S, El1, S1)
**        - E       is a value environment;
**        - Fl      is a list of formal parameters;
**        - El, El1 are actual expressions;
**        - S, S1   are stores.
**
**    True if, in the environment E and with the list of formals Fl,
**    the actual expression configuration <El, S> transitions into
**    <El1, S1> in a number of steps.
**
**    See above for details about the actual number of steps
**    performed.
*/

/*
**    Call by value.
**    --------------
*/

ae1t(E, [ival(_, _)|Form], [C|Ae], S, [C|Ae1], S1) :-
     isConst(C),
     ae1t(E, Form, Ae, S, Ae1, S1).
ae1t(E, [ival(_, _)|_], [C|Ae], S, [C1|Ae], S1) :-
     expStar(E, C, S, C1, S1).

/*
**    Call by name.
**    -------------
*/

ae1t(E, [name(_, _)|Form], [Exp|Ae], S, [NewExp|Ae1], S1) :-
    eFV(Exp, V),
    (V=[] ->
        NewExp = Exp
    ;
        (projection(E, V, E0), NewExp = closureE(E0, Exp))),
    ae1t(E, Form, Ae, S, Ae1, S1).

/*
**    Call by reference.
**    ------------------
*/

ae1t(E, [ref(_, _)|Form], [i(X)|Ae], S, [L|Ae1], S1) :-
    getEnvV(E, X, H),
    (H=l(L) ; H=L),
    ae1t(E, Form, Ae, S, Ae1, S1).

ae1t(E, [ref(_, _)|Form], [aref(X, El)|Ae], S, [L|Ae1], S1) :-
    isConstList(El),
    getEnvV(E, X, a(Limits, Base)),
    arrayIndex(Limits, El, Offset),
    L is Base+Offset,
    ae1t(E, Form, Ae, S, Ae1, S1).

ae1t(E, [ref(_, _)|_], [aref(X, El)|Ae], S, [aref(X, El1)|Ae], S1) :-
    el1t(E, El, S, El1, S1).

/*
**    Call by constant.
**    -----------------
*/

ae1t(E, [const(_, _)|Form], [C|Ae], S, [C|Ae1], S1) :-
     isConst(C),
     ae1t(E, Form, Ae, S, Ae1, S1).
ae1t(E, [const(_, _)|_], [C|Ae], S, [C1|Ae], S1) :-
     expStar(E, C, S, C1, S1).

/*
**    Call by value-result (copy in - copy out).
**    ------------------------------------------
*/

ae1t(E, [copy(_, _)|Form], [i(X)|Ae], S, [lr(L, R)|Ae1], S1) :-
    getEnvV(E, X, l(L)),
    getStore(S, L, R),
    ae1t(E, Form, Ae, S, Ae1, S1).

ae1t(E, [copy(_, _)|_], [C|Ae], S, [C1|Ae], S1) :-
    expStar(E, C, S, C1, S1).

/*
**    Functional parameters.
**    ----------------------
*/

ae1t(E, [funcpar(_, _, _)|Form], [func(F, B)|Ae], S, [func(F, B)|Ae1], S1) :-
    ae1t(E, Form, Ae, S, Ae1, S1).
ae1t(E, [funcpar(_, _, _)|Form], [i(FuncId)|Ae], S, [func(F, B)|Ae1], S1) :-
    getEnvV(E, FuncId, func(F, B)),
    ae1t(E, Form, Ae, S, Ae1, S1).

/*
**    Procedural parameters.
**    ----------------------
*/

ae1t(E, [procpar(_, _)|Form], [proc(F, B)|Ae], S, [proc(F, B)|Ae1], S1) :-
    ae1t(E, Form, Ae, S, Ae1, S1).
ae1t(E, [procpar(_, _)|Form], [i(ProcId)|Ae], S, [proc(F, B)|Ae1], S1) :-
    getEnvV(E, ProcId, proc(F, B)),
    ae1t(E, Form, Ae, S, Ae1, S1).

ae1t(_, [], [], S, [], S).

/*
**    isConstList(El)
**        - El is a list of expressions.
**
**    True if El is a list of constants.
*/

isConstList([C | Ac]) :-
    isConst(C),
    isConstList(Ac).
isConstList([]).

/*
**    comTrace(E, C, S, Cf, Sf)
**        - E     is a value environment;
**        - C, Cf are commands;
**        - S, Sf are stores.
**
**    True if, in the environment E, the command configuration <C, S>
**    transitions into the (terminal) configuration <Cf, Sf>.
**    As a side-effect, the evaluation of the command is traced.
*/

comTrace(_, halt, S, halt, S) :- !,
        outputConfig(halt),
        outputStore(S).

comTrace(E, C, S, Cf, Sf) :-
        outputConfig(C),
        outputStore(S),
        c1t(E, C, S, C1, S1), !, comTrace(E, C1, S1, Cf, Sf).

/*
**    Clauses of dynamic semantics that depend on the tracer status.
**    ==============================================================
*/

/*
**    expStar(A, E, S, E1, S1)
**        - A     is a value environment;
**        - E, E1 are expressions;
**        - S, S1 are stores.
**
**    True if, in the environment A, the expression configuration
**    <E, S> transitions into the configuration <E1, S1>.
**    If the tracer is active, expStar/5 is equivalent to e1t/5;
**    otherwise expStar/5 invokes itself recursively until E1 is a
**    constant.
*/

traceDependentClause(on, (
    expStar(E, X, S, X1, S1) :- e1t(E, X, S, X1, S1), !
)).

traceDependentClause(off, (
    expStar(_, X, S, X, S) :- isConst(X), !
)).

traceDependentClause(off, (
    expStar(E, X, S, Y, Sf) :-
        e1t(E, X, S, X1, S1), !,
        expStar(E, X1, S1, Y, Sf)
)).

/*
**    declStar(E, D, S, D1, S1)
**        - E     is a value environment;
**        - D, D1 are declarations;
**        - S, S1 are stores.
**
**    True if in the environment E the declaration configuration
**    <D, S> transitions into the configuration <D1, S1>.
**    If the tracer is active, declStar/5 is equivalent to d1t/5;
**    otherwise declStar/5 invokes itself recursively until D1 is an
**    environment.
*/

traceDependentClause(on, (
    declStar(E, D, S, D1, S1) :- d1t(E, D, S, D1, S1), !
)).

traceDependentClause(off, (
    declStar(_, X, S, X, S) :- isEnv(X), !
)).

traceDependentClause(off, (
    declStar(E, X, S, Y, Sf) :-
        d1t(E, X, S, X1, S1), !,
        declStar(E, X1, S1, Y, Sf)
)).

/*
**    comStar(E, C, S, C1, S1)
**        - E     is a value environment;
**        - C, C1 are commands;
**        - S, S1 are stores.
**
**    True if, in the environment E, the command configuration
**    <C, S> transitions into the configuration <C1, S1>.
**    If the tracer is active, comStar/5 is equivalent to c1t/5;
**    otherwise comStar/5 invokes itself recursively until C1 is
**    `halt' (<halt, S> is a terminal configuration for commands).
*/

traceDependentClause(on, (
    comStar(E, C, S, C1, S1) :- c1t(E, C, S, C1, S1), !
)).

traceDependentClause(off, (
    comStar(_, halt, S, halt, S) :- !
)).

traceDependentClause(off, (
    comStar(E, X, S, Y, Sf) :-
        c1t(E, X, S, X1, S1), !,
        comStar(E, X1, S1, Y, Sf)
)).

/*
**    topDynamicImperative(C)
**        - C is a command.
**
**    True if there exists a (finite) sequence of transitions that, in
**    the empty environment, brings <C, []> to <halt, S>, where [] is
**    the emptystore and S is any store.
**
**    (Actually, we have S=[] since we dispose locations when leaving
**    blocks.)
*/

traceDependentClause(off, (
    topDynamicImperative(C) :-
        comStar([], C, [], halt, _)
)).

traceDependentClause(on, (
    topDynamicImperative(C) :-
        comTrace([], C, [], halt, _)
)).

/*
**    getEnvV(E, X, V)
**        - E represents an SIL run-time environment;
**        - X is an identifier;
**        - V is a denotable value.
**
**    True if, in the environment E, X is associated to V.
*/

getEnvV(E, X, V) :-
    getEnv(E, X, d(_, V)).
