/* The tokenizer for CLAIR's SIL and SFL languages.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `tokenize'
**
**    This module contains the tokenizer for the SIL and SFL languages.
**    The predicates defined are significantly modified versions of
**    the ones described in the book by Clocksin & Mellish.
**
**    Predicates defined in this module:
**
**    doubleCharacter/2, inToken/2, moreString/4, readString/3,
**    readString/4, readToken/3, singleCharacterToken/1, skipComment/0,
**    skipComment1/0, startComment/2, tokenRemainder/3, tokenize/2,
**    tokenizeProgram/1. 
*/

/*
**    tokenizeProgram(L)
**
**    This is the tokenizer's main predicate.
**    L is instantiated with the list of tokens obtained from reading
**    the current input stream up to EOF.
*/

tokenizeProgram(L) :- get_code(C), tokenize(C, L).

% Stop tokenization at end of file, i.e., when get_code(-1).
tokenize(-1, []) :-
    !.

tokenize(C, X) :-
    readToken(C, Word, C1),
    ((Word = skip) -> X = Words ; X = [Word|Words]),
    tokenize(C1, Words).

readToken(34, W, C) :- !,
    readString(L, 34, C), name(W, L).

readToken(C, W, C2) :- singleCharacterToken(C),
    !,
    get_code(C1),
    (startComment(C, C1) ->
        (skipComment(End), !, W=skip, (End = ended -> get_code(C2); C2 = -1))
   ;
        (doubleCharacter(C, C1) ->
            (name(W, [C, C1]), get_code(C2))
        ;
            (name(W, [C]), C2 = C1))).

readToken(C, W, C2) :-
    inToken(C, NewC),
    !,
    get_code(C1),
    tokenRemainder(C1, Cs, C2),
    name(W, [NewC|Cs]).

readToken(_, skip, C1) :-
    get_code(C1).

tokenRemainder(C, [NewC|Cs], C2) :-
    inToken(C, NewC),
    !,
    get_code(C1),
    tokenRemainder(C1, Cs, C2).
tokenRemainder(C, [], C).

readString(Chars, Quote, NextCh) :-
    get_code(Ch),
    readString(Ch, Chars, Quote, NextCh).

readString(-1, _, Quote, -1) :-
    write(' error: end of file in '),
    put_code(Quote),
    write(token),
    put_code(Quote),
    nl,
    !,
    fail.
readString(Quote, Chars, Quote, NextCh) :- !,
    get_code(Ch),                               /* Closed or double quote */
    moreString(Ch, Quote, Chars, NextCh).
readString(Char, [Char|Chars], Quote, NextCh) :-
    readString(Chars, Quote, NextCh).       /* Normal character */

moreString(Quote, Quote, [Quote|Chars], NextCh) :- !,
    readString(Chars, Quote, NextCh).       /* Double quote */
moreString(NextCh, _, [], NextCh).          /* End */

startComment(47, 42).

skipComment(End) :-
    get_code(C),
    ((C = -1, End = endless) ; 
     (C=42, skipComment1(End)) ; 
     skipComment(End)
    ).


skipComment1(End) :-
    get_code(C),
    ((C = -1, End = endless) ;
     (C=47, End = ended) ;
     (C=42, skipComment1(End)) ; 
     skipComment(End)
    ).

singleCharacterToken(40).  /* ascii('(', 40)  */
singleCharacterToken(41).  /* ascii(')', 41)  */
singleCharacterToken(42).  /* ascii('*', 42)  */
singleCharacterToken(43).  /* ascii('+', 43)  */
singleCharacterToken(44).  /* ascii(',', 44)  */
singleCharacterToken(45).  /* ascii('-', 45)  */
singleCharacterToken(46).  /* ascii('.', 46)  */
singleCharacterToken(47).  /* ascii('/', 47)  */
singleCharacterToken(58).  /* ascii(':', 58)  */
singleCharacterToken(59).  /* ascii(';', 59)  */
singleCharacterToken(60).  /* ascii('<', 60)  */
singleCharacterToken(61).  /* ascii('=', 61)  */
singleCharacterToken(62).  /* ascii('>', 62)  */
singleCharacterToken(91).  /* ascii('[', 91)  */
singleCharacterToken(93).  /* ascii(']', 93)  */
singleCharacterToken(124). /* ascii('|', 124) */

doubleCharacter(58, 61). /* := */
doubleCharacter(61, 61). /* == */
doubleCharacter(60, 61). /* <= */
doubleCharacter(62, 61). /* >= */
doubleCharacter(60, 62). /* <> */
doubleCharacter(46, 46). /* .. */
doubleCharacter(45, 62). /* -> */

inToken(C, C) :- C >= 97, C =< 122 .  /* the lower-case letters */
inToken(C, C) :- C >= 65, C =< 90 .   /* the upper-case letters */
inToken(C, C) :- C >= 48, C =< 57 .   /* the digits */
inToken(95, 95). /* ascii('_', 95) */
